
;--- DOS device driver, emulates a hole in int 15h, ax=e820h.
;--- Public Domain, written by japheth.
;--- create with: jwasm -mz i15e820.asm

	.286
	.MODEL SMALL
	.dosseg
	.386

	.stack 400h

;*** macros and structures

IODAT   struc
cmdlen  db      ?       ;+ 0:size
unit    db      ?       ;+ 1:
cmd     db      ?       ;+ 2
status  dw      ?       ;+ 3
        db      8 dup (?); reserved
media   db      ?       ;+ 0d
trans   dd      ?       ;+ 0e
count   dw      ?       ;+ 12   init:offset parameter line
start   dw      ?       ;+ 14   init:segment parameter line
drive   db      ?       ;+ 16
IODAT   ends

E820MAP struct
baselow dd  ?
basehigh dd  ?
lenlow  dd  ?
lenhigh dd  ?
type_   dd  ?
E820MAP ends

CStr macro text:vararg
local xxx
	.const
xxx db text, 0
	.code
	exitm <offset xxx>
endm


	.CODE

	dd -1
	dw 8000h				  ;attribute
	dw offset devstrat		  ;device strategy
	dw offset devint		  ;device interrupt
devname db 'I15E820$'

cmdptr	dd 1 dup (?)

devstrat proc far
	mov cs:word ptr[cmdptr+0],bx
	mov cs:word ptr[cmdptr+2],es
	ret
devstrat endp


devint proc far
	pusha
	push ds
	push es
	lds bx,cs:[cmdptr]
	mov word ptr [bx.IODAT.status],100h
	mov al,[bx.IODAT.cmd]
	cmp al,00			;init call?
	jnz @F
	mov word ptr [bx.IODAT.trans+0],offset resident
	mov word ptr [bx.IODAT.trans+2],cs
	call install
@@:
	pop es
	pop ds
	popa
	ret
devint endp

blocklen dd -1
id       dd -1
len      dd -1
emode    db 0
	align 2

myint15 proc far
	pushf
	cmp ax, 0E820h
	jz itsme
	popf
	db 0EAh
nextd::
	dd 0
itsme:
	cmp cs:[emode],0
	jnz xmode
	clc
	call dword ptr cs:[nextd]
	jc error
	cmp eax, "SMAP"
	jnz error
	cmp ebx,0
	jz done
	cmp es:[di].E820MAP.basehigh,0
	jnz done
	cmp es:[di].E820MAP.lenhigh,0
	jnz done
	cmp es:[di].E820MAP.type_,1	;free mem?
	jnz done
	cmp es:[di].E820MAP.baselow,100000h	;1 MB
	jnz done
	push eax
	mov eax, es:[di].E820MAP.lenlow
	mov cs:[blocklen],eax
	mov cs:[id],ebx
	mov cs:[len],ecx
	mov cs:[emode],1
	shr eax, 1
	mov es:[di].E820MAP.lenlow, eax
	pop eax
done:
	push bp
	mov bp,sp
	and byte ptr [bp+6],0feh
	pop bp
	iret
error:
	push bp
	mov bp,sp
	or byte ptr [bp+6],1
	pop bp
	iret

xmode:
	popf
	mov ebx,cs:[id]
	mov eax,"SMAP"
	mov ecx,cs:[len]
	cmp cs:[emode],1
	jnz xmode2
	push eax
	mov eax,cs:[blocklen]
	shr eax,1
	add eax, 100000h
	mov es:[di].E820MAP.baselow, eax
	mov es:[di].E820MAP.lenlow, 1000h
	mov es:[di].E820MAP.type_, 2
	pop eax
	inc cs:[emode]
	jmp done
xmode2:
	push eax
	mov eax,cs:[blocklen]
	shr eax,1
	add eax, 101000h
	mov es:[di].E820MAP.baselow, eax
	mov eax,cs:[blocklen]
	shr eax,1
	sub eax, 1000h
	mov es:[di].E820MAP.lenlow, eax
	mov es:[di].E820MAP.type_, 1
	pop eax
	mov cs:[emode],0
	jmp done

myint15 endp

resident equ $

install proc

	mov ax,3515h
	int 21h
	mov word ptr cs:[nextd+0],bx
	mov word ptr cs:[nextd+2],es
	push ds
	mov dx, offset myint15
	push cs
	pop ds
	mov ax,2515h
	int 21h
	mov ax,@data
	mov ds,ax
	mov dx,CStr("I15E820 driver installed",13,10,'$')
	mov ah,9
	int 21h
	pop ds
	ret
install endp

main proc
	mov ax,@data
	mov ds,ax
	mov dx,CStr("must be installed as device driver in CONFIG.SYS, BEFORE the XMM",13,10,'$')
	mov ah,9
	int 21h
	mov ax,4C00h
	int 21h
main endp

	END main

